#! /usr/bin/env python3
# 
# Show what the BERT model representation is for a pair of sentences. This is based on the following:
#    https://github.com/google-research/bert/blob/master/README.md
# This includes an option to compute cosine distance matrix for the tokens in the feature vectors, This follows the approach from the following article:
#    https://medium.com/thelocalminima/are-bert-features-interbertible-250a91eb9dc
#
# Sample input (sentences 1 & 2):
#
#    "april july"
#    "spring summer"
#
# Sample output:
#
#    - Cosine matrix:
#
#            [CLS]   april   july    [SEP]   spring  summer  [SEP]
#    [CLS]   1.0    *0.344   0.336   -0.025  0.34    0.341   0.011
#    april           1.0     0.514   0.047  *0.606   0.414   0.062
#    july                    1.0     0.045   0.366   0.683   0.054
#    [SEP]                           1.0     0.094   0.064  *0.908
#    spring                                  1.0    *0.599   0.107
#    summer                                          1.0     0.082
#    [SEP]                                                  *1.0
#
#    - Feature representation:
#
#    {
#      "linex_index" : 0,
#      "features" : [
#         {
#            "layers" : [
#               {
#                  "index" : -1
#                  "values" : [
#                     -0.460937, 0.192811, -0.473121, -0.254216, 0.077836, -0.139587, 0.613976, 0.932992, ...
#                  ],
#               },
#               {
#                  "index" : -2
#                  "values" : [
#                     -0.89061, -0.080438, -0.519494, -0.280097, 0.936842, -0.11848, -0.204586, 1.619458, ...
#                  ],
#               },
#               {
#                  "index" : -3
#                  "values" : [
#                     -0.83093, 0.116799, -0.590249, -0.312695, 0.330304, -0.166485, 0.380528, 1.59396, ...
#                  ],
#               },
#               {
#                  "index" : -4,
#                  "values" : [
#                     -0.833253, 0.129633, -0.721126, -0.315408, 0.435143, 0.018215, 0.090608, 0.985256, ...
#                  ]
#               }
#            ],
#            "token" : "[CLS]"
#         },
#         {
#            "token" : "april",
#            "layers" : [
#               {
#                  "index" : -1
#                  "values" : [
#                     -0.86163, -0.03601, -0.080842, -0.166804, -0.272864, -0.062401, 0.226085, 0.556695, ...
#                  ],
#               },
#               {
#                  "index" : -2
#                  "values" : [
#                     -0.643406, -0.383428, -0.145636, 0.390536, -0.157737, -0.209778, 0.100278, 0.810878, ...
#                  ],
#               },
#               {
#                  "index" : -3,
#                  "values" : [
#                     -0.568553, -0.327431, 0.032729, 0.634031, -0.089157, -0.165046, 0.454916, 0.643462, ...
#                  ]
#               },
#               {
#                  "index" : -4,
#                  "values" : [
#                     -0.428569, -0.138098, -0.508629, 0.325682, -0.215687, -0.293736, 0.453519, 0.44138, ...
#                  ]
#               }
#            ]
#         },
#         ...
#    }
#
#................................................................................
# Notes:
# - To get context-insensitive vectors, the script should be run separately
#   for each word.
# - To show cosine similarity for complex vector terms (e.g., king - man + woman),
#   get the individual work vectors and manually combine vectors (e.g., ipython)
#   before invoking show_cosine_distances.
#................................................................................
# TOOD:
# - Add option to average the vectors rather than using last (i.e., index 3).
#

"""Show the representation of a BERT model for a sentence or pair (see BERT README.md)"""

# Standard packages
import json
import os
import re
import sys

# Installed packages
import numpy as np
import scipy.spatial.distance
try:
    # pylint: disable=import-error
    tf = None
    if ("--help" not in sys.argv):
        import tensorflow as tf
except:
    ## TODO: system.print_exception_info("tensorflow import")
    sys.stderr.write(f"Problem importing tensorflow: {sys.exc_info()}")
if not tf:
    ## TODO: tf = object(); move into supporting module (e.g., misc_utils)
    class Fubar:
        """Fouled-up, etc."""
        pass
    f_inner = Fubar()
    tf = Fubar()
    ## TODO: tf = json.loads('{"version": {"VERSION": "-1"}}')
    setattr(f_inner, "VERSION", "-1")
    setattr(tf, "version", f_inner)

# Local packages
from mezcla import debug
from mezcla.main import Main
from mezcla import system
from mezcla import glue_helpers as gh
from mezcla import text_utils
from mezcla.system import form_path, getenv_bool, getenv_int, getenv_text, getenv_value, round_num
from mezcla.text_utils import version_to_number as version_as_float, make_fixed_length

# Constants
SENTINEL_TOKENS = ["[CLS]", "[SEP]"]

# Environment constants
PYTHON = getenv_text("PYTHON", "python", 
                     description="Command for invoking script (e.g., /usr/bin/nice 19 python -u)")
PROJECT_DIR = getenv_text("PROJECT_DIR", "/usr/local/misc",
                          description="Base directory for project files")
CODE_DIR = getenv_text("CODE_DIR", form_path(PROJECT_DIR, "programs", "python"),
                       description="Base directory for python scripts")
DATA_DIR = getenv_text("DATA_DIR", form_path(PROJECT_DIR, "data"),
                       description="Base directory for program data")
DEEP_LEARNING_DATA = getenv_text("DEEP_LEARNING_DATA", form_path(DATA_DIR, "deep-learning"),
                                 description="Base directory for deep learning data")
USE_ALBERT = getenv_bool("USE_ALBERT", False,
                         description="Use ALBERT instead of BERT")
## TODO: PRE = "al" if USE_ALBERT else ""
BERT = "albert" if USE_ALBERT else "bert"
BERT_CODE_DIR = getenv_text("BERT_CODE_DIR", form_path(CODE_DIR, BERT),
                            description="Code for BERT specialization (e.g., ALBERT)")
DEFAULT_BERT_DATA_DIR = "albert_base" if USE_ALBERT else "uncased_L-12_H-768_A-12"
BERT_DATA_DIR = getenv_text("BERT_DATA_DIR", 
                            form_path(DEEP_LEARNING_DATA, BERT, DEFAULT_BERT_DATA_DIR),
                            description="Directory for pre-trained data")
BERT_MODEL_DIR = getenv_text("BERT_MODEL_DIR", BERT_DATA_DIR,
                             "Directory for BERT models which can differ from BERT_DATA_DIR if pre-training")
## OLD: DEFAULT_CHECKPOINT_FILE = "model.ckpt" if USE_ALBERT else "bert_model.ckpt"
## TODO: add Python helper to do this via glob.glob, etc.:
# note: Uses the latest checkpoint file in the BERT model directory
DEFAULT_CHECKPOINT_FILE = (gh.run(f"ls -t {BERT_MODEL_DIR}/*ckpt*.index | head -1").replace(".index", ""))
CHECKPOINT_FILE = getenv_text("CHECKPOINT_FILE", DEFAULT_CHECKPOINT_FILE,
                              description="BERT pre-taining model checkpoint")
MAX_SEQ_LENGTH = getenv_value("MAX_SEQ_LENGTH", None,
                             "Maximum sentence token sequence length")
BATCH_SIZE = getenv_value("BATCH_SIZE", None,
                         "Batch size for processing")
NUM_LAYERS = getenv_int("NUM_LAYERS", 4,
                        "Number of layers to extract")
AVERAGE_LAYERS = getenv_bool("AVERAGE_LAYERS", False,
                             "Take average of layers rather than last one")
FIELD_WIDTH = getenv_int("FIELD_WIDTH", None,
                         "Width of fields for cosine matrix")

# Option names
MODEL = "model"
SENTENCE1 = "sentence1"
SENTENCE2 = "sentence2"
COSINE = "cosine"
VECTORS = "vectors"
LAYER = "layer"
ALL_TOKENS = "all-tokens"
VERBOSE = "verbose"
FILTER = "filter"
INCLUDE = "include"

def cosine_distance(vector1, vector2):
    """Compute cosine distance between VECTOR1 and VECTOR2"""
    # EX: cosine_distance([1, 0, 0], [0, 0, 1]) => 1.0
    # EX: cosine_distance([1, 0, 0], [2, 0, 0]) => 0.0
    # EX: cosine_distance([1, 0, 0, 0], [1, 1, 1, 1]) => 0.5
    dist = scipy.spatial.distance.cosine(vector1, vector2)
    debug.trace_fmt(5, "cosine_distance({v1}, {v2}) => {d})",
                    v1=vector1, v2=vector2, d=dist)
    return dist


def print_vector(vector, fixed_width=None):
    """Print VECTOR tab-delimited unless FIXED_WIDTH"""
    debug.assertion(all(isinstance(value, str) for value in vector))
    if fixed_width:
        print("".join(make_fixed_length(value, fixed_width)
                      for value in vector))
    else:
        print("\t".join(vector))
    return


def show_cosine_distances(tokens, vectors):
    """Displays the cosine distance matrix (all pairs of tokens)
    Note: prints entries tab-separated unless FIELD_WIDTH global set
    """

    # Calculate distances and display matrix (omitting symmetric cases)
    print_vector([""] + tokens, fixed_width=FIELD_WIDTH)
    for i, token1 in enumerate(tokens):
        info = [token1] + ([""] * i)
        for j in range(i, len(tokens)):
            dist = cosine_distance(vectors[i], vectors[j])
            info.append(str(round_num(dist, 3)))
        print_vector(info, fixed_width=FIELD_WIDTH)
    return


class ExtractFeatures(object):
    """Class for extracting features from BERT model"""

    def __init__(self, model=None):
        """Class constructor with optional MODEL override"""
        if model is None:
            model = CHECKPOINT_FILE
        self.model = model
        debug.assertion(system.file_exists(self.model + ".index"))
        return

    def run(self, sentence1, sentence2, basename=None, other_options=None):
        """Invoke the feature extraction and return JSON result as string"""
        # Set options
        if basename is None:
            basename = "_features"
        if other_options is None:
            other_options = ""
        model_path = self.model
        if not system.file_exists(model_path + ".index"):
            model_path = system.form_path(BERT_MODEL_DIR, self.model)

        # Create input file in correct format from sentences
        SENTENCE_DELIM = "|||"
        debug.assertion(SENTENCE_DELIM not in (sentence1 + sentence2))
        input_filename = basename + ".txt"
        input_text = sentence1
        if sentence2.strip():
             input_text += f" {SENTENCE_DELIM} {sentence2}"
        system.write_file(input_filename, input_text)
        output_filename = basename + ".json"

        # Invoke the feature extraction script
        # TODO: document the options; optionally, include all layers
        D = os.path.sep
        layer_spec = ",".join(str(-i) for i in range(1, NUM_LAYERS + 1))
        command_line = f"""{PYTHON} {BERT_CODE_DIR}{D}extract_features.py  {other_options} \
                       --input_file={input_filename} \
                       --output_file={output_filename} \
                       --vocab_file={BERT_DATA_DIR}{D}vocab.txt \
                       --bert_config_file={BERT_DATA_DIR}{D}{BERT}_config.json \
                       --init_checkpoint={model_path} \
                       --layers={layer_spec}"""
        if MAX_SEQ_LENGTH:
            command_line += f"--max_seq_length={MAX_SEQ_LENGTH}"
        if BATCH_SIZE:
            command_line += f"--batch_size={BATCH_SIZE}"
        gh.run(command_line)
        # TODO: check for common errors (e.g., mismatch in -max_seq_length_

        # Read the JSON representation
        # TODO: pretty print if necessary; add option to prune
        json_representation = system.read_file(output_filename)
        debug.assertion(re.search(r"{.*}", json_representation, re.DOTALL))

        # Cleanup
        if basename and not debug.detailed_debugging():
            # TODO: add helper for this in system
            map(gh.delete_file, gh.get_matching_files(basename + "*"))

        # Return string with features
        return json_representation


class Script(Main):
    """Adhoc script class (e.g., no I/O loop, just run calls)"""
    model = None
    sentence1 = None
    sentence2 = None
    calc_cosine = False
    show_vectors = False
    layer = 3
    use_all_tokens = False
    verbose = False
    include = False
    filter_terms = ""

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.model = self.get_parsed_option(MODEL, self.model)
        self.sentence1 = self.get_parsed_argument(SENTENCE1)
        self.sentence2 = self.get_parsed_argument(SENTENCE2)
        self.layer = self.get_parsed_option(LAYER, self.layer)
        self.calc_cosine = self.get_parsed_option(COSINE, self.calc_cosine)
        self.show_vectors = self.get_parsed_option(VECTORS, self.show_vectors)
        self.use_all_tokens = self.get_parsed_option(ALL_TOKENS, self.use_all_tokens)
        self.verbose = self.get_parsed_option(VERBOSE, self.verbose)
        self.include = self.get_parsed_option(INCLUDE, self.include)
        self.filter_terms = self.get_parsed_option(FILTER, self.filter_terms)
        debug.trace_object(5, self, label="Script instance")
        return

    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)

        # Make sure number of requested layers doesn't exceed model size
        bert_config_file = gh.form_path(BERT_DATA_DIR, BERT + "_config.json")
        bert_config = json.loads(system.read_file(bert_config_file))
        global NUM_LAYERS
        num_actual_layers = bert_config.get("num_hidden_layers", 1)
        if num_actual_layers < NUM_LAYERS:
            # TODO: only show warning id NUM_LAYERS not the default (4)
            debug.trace(3, f"Warning: only {num_actual_layers} available so not getting {NUM_LAYERS}")
            NUM_LAYERS = num_actual_layers

        # Invoke the helper class to do the actual invocation
        extractor = ExtractFeatures(model=self.model)
        ## OLD: json_repr = extractor.run(self.sentence1, self.sentence2, basename=self.temp_file)
        json_repr = extractor.run(self.sentence1, self.sentence2, basename=self.temp_base)
        tokens = vectors = []
        show_repr = True
        if self.show_vectors or self.calc_cosine:
            (tokens, vectors) = self.extract_vectors(json_repr)
            debug.assertion(len(tokens) == len(vectors))
            show_repr = False
        if self.show_vectors:
            for i, vector in enumerate(vectors):
                if self.verbose:
                    print(tokens[i] + ": ")
                print(vector)
        if self.calc_cosine:
            show_cosine_distances(tokens, vectors)
        if show_repr:
            print(json_repr)
        return

    def extract_vectors(self, json_repr):
        """Extract vectors for each token (from self.layer)"""
        # Filter features/tokens (normally to exclude special ones or to match user list)
        feature_info = json.loads(json_repr)
        all_features = feature_info["features"]
        all_tokens = [h["token"] for h in all_features]
        tokens = all_tokens if self.use_all_tokens else system.difference(all_tokens, SENTINEL_TOKENS)
        if self.filter_terms:
            terms_to_filter = text_utils.extract_string_list(self.filter_terms)
            filter_fn = (system.intersection if self.include else system.difference)
            tokens = filter_fn(tokens, terms_to_filter)
        tokens = sorted(tokens)
        features = [h for (i, h) in enumerate(all_features) if all_tokens[i] in tokens]
        debug.assertion(all((len(h["layers"]) == NUM_LAYERS) for h in features))

        # Either average vectors or take last
        if AVERAGE_LAYERS:
            vectors = []
            for h in features:
                vector = np.array(h["layers"][0]["values"])
                for layer in range(1, NUM_LAYERS):
                    vector += np.array(h["layers"][layer]["values"])
                vector = vector / NUM_LAYERS
                vectors.append(list(vector))
        else:
            vectors = [h["layers"][self.layer]["values"] for h in features]
        return (tokens, vectors)

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    app = Script(
        description=__doc__,
        # Note: skip_input controls the line-by-line processing, which is inefficient but simple to
        # understand; in contrast, manual_input controls iterator-based input (the opposite of both).
        skip_input=True,
        manual_input=True,
        boolean_options=[(ALL_TOKENS, "Include all tokens (e.g., sentinels)"),
                         (COSINE, "Show cosine between vectors"),
                         (VECTORS, "Show feature vectors for all tokens"),
                         (INCLUDE, "Use an inclusion filter"),
                         (VERBOSE, "Verbose output mode")],
        int_options=[(LAYER, "Which layer to use (0-3)")],
        positional_options=[SENTENCE1, SENTENCE2],
        text_options=[(MODEL, "Basename of BERT model file"),
                      (FILTER, "Terms to filter from output")]
        )
    debug.assertion(version_as_float("1.13") <= version_as_float(tf.version.VERSION) < version_as_float("2.0"))
    app.run()



#! /usr/bin/env python3
#
# Simple illustration of optional pydantic argument validation.
#

"""Optional pydantic argument validation"""

# Standard modules
import re
from typing import Optional
import importlib.util
import sys

# Installed module
from pydantic import validate_call

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system
from mezcla.my_regex import my_re

def validate_script_args(script_path):
    """Dynamical validation of arguments of functions of the specified script"""
    try:
        spec = importlib.util.spec_from_file_location("script_module", script_path)
        script_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(script_module)
    except Exception as e:
        print(f"Error importing script: {e}")
        return

    functions_to_validate = [
        getattr(script_module, func_name)
        for func_name in dir(script_module)
        if callable(getattr(script_module, func_name))
        and hasattr(getattr(script_module, func_name), "__annotations__")
    ]

    print(functions_to_validate)

    for func in functions_to_validate:
        try:
            validated_func = validate_call(func)
            print(f"Validating arguments for function: {func.__name__}")
            validated_func()
        except Exception as e:
            print(f"Error validating arguments for function {func.__name__}: {e}")

def main():
    """Entry point"""
    if len(sys.argv) < 2:
        print("Usage: python validate_script_arguments.py <path_to_script>")
        sys.exit(1)

    script_path = sys.argv[1]
    validate_script_args(script_path)

if __name__ == '__main__':
    main()





#! /usr/bin/env python3
#
# Test(s) for ../data_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_data_utils.py
#

"""Tests for data_utils module"""

# Standard packages
import os
import pandas as pd

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.data_utils as THE_MODULE

class TestDataUtils(TestWrapper):
    """Class for testcase definition"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    path = os.path.dirname(os.path.realpath(__file__))

    def test_read_csv(self):
        """Ensure read_csv works as expected"""
        debug.trace(4, "test_read_csv()")
        tf = THE_MODULE.read_csv(f"{self.path}/../examples/iris.csv")
        assert tf.shape == (150, 5)

    def test_to_csv(self):
        """Ensure to_csv works as expected"""
        debug.trace(4, "test_to_csv()")
        system.setenv("DELIM", ",")
        # Setup
        temp_file = gh.get_temp_file()
        df = pd.DataFrame()
        df['sepal_length'] = [5.1, 4.9, 4.7, 4.6, 5.0]
        df['sepal_width'] = [3.5, 3.0, 3.2, 3.1, 3.6]
        df['petal_length'] = [1.4, 1.4, 1.3, 1.5, 1.4]
        df['petal_width'] = [0.2, 0.2, 0.2, 0.2, 0.2]
        df['class'] = ['Iris-setosa', 'Iris-virginica', 'Iris-versicolor', 'Iris-setosa', 'Iris-setosa']
        # TODO2: track down exception
        #  Exception during write_csv: (<class 'TypeError'>, TypeError('"delimiter" must be string, not NoneType'), <traceback object at 0x7f49885f3880>)
        THE_MODULE.to_csv(temp_file, df)
        df.to_csv(temp_file, index=False)
        # Test
        expected = (
            'sepal_length,sepal_width,petal_length,petal_width,class'
        )
        assert expected in gh.read_file(temp_file)

    def test_lookup_df_value(self):
        """Ensure lookup_df_value works as expected"""
        debug.trace(4, "test_lookup_df_value()")
        tf = THE_MODULE.read_csv(f"{self.path}/../examples/iris.csv")
        assert THE_MODULE.lookup_df_value(tf, "sepal_length", "petal_length", "3.8") == "5.5" 

    def test_main(self):
        """Ensure main works as expected"""
        debug.trace(4, "main()")
        THE_MODULE.main()
        captured = self.get_stderr()
        assert "Error" in captured

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Test(s) for ../text_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_text_utils.py
#

"""Tests for text_utils module"""

# Standard packages
import re
import os

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system
from mezcla.unittest_wrapper import TestWrapper, invoke_tests

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.text_utils as THE_MODULE

HTML_FILENAME = "./resources/simple-window-dimensions.html"

EXPECTED_TEXT = """
   Simple window dimensions

   Simple window dimensions

      Legend:
        Screen dimensions:    ???
        Browser dimensions:   ???

   JavaScript is required
"""
#
# NOTE: Whitespace and punctuation gets normalized
# TODO: restore bullet points (e.g., "* Screen dimensions")

MS_WORD_FILENAME = "./resources/spanish-accents.docx"

MS_WORD_TEXT = "Tío Tomás\t\t\t\tUncle Tom\n\n¡Buenos días!\t\t\t\tGood morning\n\nçãêâôöèäàÃëÇÂîòïÔìðÊÅåùÀŠý\t\tcaeaooeaaAeCAioiOioEAauASy"


def normalize_text(text):
    """Trim excess whitespace and convert punctuation to <PUNCT>"""
    # EX: normalize_test_text("   h  e y?! ") => "h e y<PUNCT>"
    result = text.strip()
    result = re.sub(r"\s+", " ", result)
    result = re.sub(r"[^\w\s]+", "<PUNCT>", result)
    debug.trace(4, f"normalize_test_text({text}) => {result}")
    return result


class TestTextUtils(TestWrapper):
    """Class for test case definitions"""
    # note: script_module used in argument parsing sanity check (e.g., --help)
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def test_init_BeautifulSoup(self):
        """Ensure init_BeautifulSoup works as expected"""
        debug.trace(4, "test_init_BeautifulSoup()")
        THE_MODULE.BeautifulSoup = None
        THE_MODULE.init_BeautifulSoup()
        assert THE_MODULE.BeautifulSoup is not None

    def test_html_to_text(self):
        """Ensure html_to_text works as expected"""
        # TODO: move into test_html_utils.py
        debug.trace(4, "test_html_to_text()")
        html_path = gh.resolve_path(HTML_FILENAME)
        html = system.read_file(html_path)
        text = THE_MODULE.html_to_text(html)
        assert normalize_text(text) == normalize_text(EXPECTED_TEXT)

    def test_init_textract(self):
        """Ensure init_textract works as expected"""
        debug.trace(4, "test_init_textract()")
        THE_MODULE.textract = None
        THE_MODULE.init_textract()
        assert THE_MODULE.textract is not None

    def test_document_to_text(self):
        """Ensure document_to_text works as expected"""
        debug.trace(4, "test_document_to_text()")
        doc_path = gh.resolve_path(MS_WORD_FILENAME)
        text = THE_MODULE.document_to_text(doc_path)
        assert normalize_text(text) == normalize_text(MS_WORD_TEXT)

    def test_extract_html_images(self):
        """Ensure extract_html_images works as expected"""
        debug.trace(4, "test_extract_html_images()")

        url = 'example.com'
        html = (
            '<!DOCTYPE html>\n'
            '<html>\n'
            '<body>\n'
            '<h2>The target Attribute</h2>\n'
            '<div class="some-class">this is a div</div>\n'
            '<div class="some-class another-class">'
            '<img src="smiley.gif" alt="Smiley face" width="42" height="42" style="border:5px solid black">\n'
            '<img src="some_image.jpg" alt="Some image" width="42" height="42" style="border:5px solid black">\n'
            '<img src="hidden.jpg" alt="this is a hidden image" width="42" height="42" style="display:none">\n'
            '</div>'
            '</body>\n'
            '</html>\n'
        )
        images_urls = [
            f'{url}/smiley.gif',
            f'{url}/some_image.jpg'
        ]

        result = THE_MODULE.extract_html_images(html, url)
        assert result == images_urls
        assert 'hidden.jpg' not in images_urls

    def test_version_to_number(self):
        """Ensure version_to_number works as expected"""
        debug.trace(4, "test_version_to_number()")

        # Testing simple version text
        assert THE_MODULE.version_to_number("1") == 1
        assert THE_MODULE.version_to_number("") == 0
        assert THE_MODULE.version_to_number("1.11.1") == 1.00110001

        # Testing version with non-numeric components and with new line
        assert THE_MODULE.version_to_number("3.7.6 (default, Jan  8 2020, 19:59:22)<NL>[GCC 7.3.0]\n") == 3.00070006

        # Testing max padding with long and short versions
        assert THE_MODULE.version_to_number("8.12.10143", max_padding=2) == 8.012999
        assert THE_MODULE.version_to_number("15.4.9", max_padding=6) == 15.00000040000009

    def test_extract_string_list(self):
        """Ensure extract_string_list works as expected"""
        debug.trace(4, "test_extract_string_list()")

        # Testing whitespace and commas
        assert THE_MODULE.extract_string_list("") == []
        assert THE_MODULE.extract_string_list(" ") == []
        assert THE_MODULE.extract_string_list(",,") == []
        assert THE_MODULE.extract_string_list("1, 2,3 4") == ['1', '2', '3', '4']
        assert THE_MODULE.extract_string_list("    one,two,   three four   ") == ['one', 'two', 'three', 'four']

        # Testing embedded spaces with quotes
        assert THE_MODULE.extract_string_list("'my dog'  likes  'my  hot  dog'") == ['my dog', 'likes', 'my  hot  dog']
        assert THE_MODULE.extract_string_list("'my dog  likes  my  hot  dog'") == ['my dog  likes  my  hot  dog']

    def test_extract_int_list(self):
        """Ensure extract_int_list works as expected"""
        debug.trace(4, "test_extract_int_list()")

        # Testing normal delimited string
        assert THE_MODULE.extract_int_list("1, 2, 3, 4, 5") == [1, 2, 3, 4, 5]
        assert THE_MODULE.extract_int_list("1   2  3  4   5") == [1, 2, 3, 4, 5]

        # Testing DEFAULT_VALUE
        assert THE_MODULE.extract_int_list("1   2  foobar", default_value=9) == [1, 2, 9]
        assert THE_MODULE.extract_int_list("1   2  3.45", default_value=230) == [1, 2, 230]

    def test_getenv_ints(self):
        """Ensure getenv_ints works as expected"""
        debug.trace(4, "test_getenv_ints()")
        ## INFO: help(self.monkeypatch)
        self.monkeypatch.setenv("DUMMY-VARIABLE", "0, 1  2  3 4 ", prepend=os.pathsep)
        assert THE_MODULE.getenv_ints("DUMMY-VARIABLE", str(list(range(5)))) == [0, 1, 2, 3, 4]

    def test_is_symbolic(self):
        """Ensure is_symbolic works as expected"""
        debug.trace(4, "test_is_symbolic()")
        assert THE_MODULE.is_symbolic("PI")
        assert not THE_MODULE.is_symbolic("3.14159")
        assert not THE_MODULE.is_symbolic("123")
        assert not THE_MODULE.is_symbolic(123)
        assert not THE_MODULE.is_symbolic(0.1)

    def test_make_fixed_length(self):
        """Ensure make_fixed_length works as expected"""
        debug.trace(4, "make_fixed_length()")
        assert THE_MODULE.make_fixed_length("fubar", 3) == "fub"
        assert THE_MODULE.make_fixed_length("fub", 5) == "fub  "

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Miscellaneous text utility functions, such as for extracting text from
# HTML and other documents.
#
# Notes:
# - Modules in other directoriess rely upon this (e.g., ~/visual-search),
#   so do global search before making significant changes.
# - This is in the spirit of quick-n-dirty (e.g., for R&D: they are packages
#   that are better suited for industrial strength code (e.g., for production).
#
# Usage examples:
#
#   from mezcla import debug
#   from mezcla import text_utils
#
#   import tensorflow
#   from mezcla.text_utils import version_to_number as version2num
#   debug.assertion(version2num(tensorflow.__version__) <= version2num("1.15"))
#
#   devlin_text = document_to_text("devlin-bert-pretraining-arxiv-1810-04805-dec19.pdf")
#   str(devlin_text).count("BERT")
#   => 117
#
# TODO:
# - Write up test suite, el tonto!.
# - Add pointer to specific packages better for production use.
# - Move HTML-specific functions into html_utils.py.
#

"""Miscellaneous text utility functions"""

# Library packages
import re
import six
import sys

# Installed packages
## HACK: temporarily make these conditional
## from bs4 import BeautifulSoup
## import textract

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import html_utils
from mezcla.my_regex import my_re
from mezcla import system
from mezcla.system import to_int

# TEMP: Placeholders for dynamically loaded modules
BeautifulSoup = None
textract = None

def init_BeautifulSoup():
    """Make sure bs4.BeautifulSoup is loaded"""
    import bs4                           # pylint: disable=import-outside-toplevel, import-error
    global BeautifulSoup
    BeautifulSoup = bs4.BeautifulSoup


def extract_soup_text_with_breaks(soup):
    """Extract text from SOUP parse, accounting for implicit newlines"""
    # TODO: def extract_soup_text(soup: bs4.Tag) -> str:
    # Based on https://stackoverflow.com/questions/30337528/make-beautifulsoup-handle-line-breaks-as-a-browser-would
    debug.trace_fmtd(6, f"extract_soup_text({soup})")
    _inline_elements = {"a", "span", "em", "strong", "u", "i", "font", "mark", "label", "s",
                        "sub", "sup", "tt", "bdo", "button", "cite", "del", "b"}
    import bs4                           # pylint: disable=import-outside-toplevel, import-error

    # Define helper function for generating text
    def _get_text(tag):
        """Helper generator"""
        # TODO: def _get_text(tag: bs4.Tag) -> Generator:
        for child in tag.children:
            if isinstance(child, bs4.Tag):
                # if the tag is a block type tag then yield new lines before after
                is_block_element = (child.name not in _inline_elements)
                if is_block_element:
                    yield "\n"
                # note: 'yield from' used due to recursive invocation
                # TODO: see whether <br> produces extraneous newlines
                yield from (["\n"] if (child.name == "br") else  _get_text(child))
                if is_block_element:
                    yield "\n"
            elif isinstance(child, bs4.NavigableString):
                yield child.string
            else:
                debug.assertion(not ("".join(_get_text(child))).strip())

    # Extract all the text
    result = "".join(_get_text(soup))    
    debug.trace_fmtd(6, f"extract_soup_text() => {result}")
    return result


def html_to_text(document_data):
    """Returns text version of html DATA
    Warning: deprecated function; import from html_utils instead
    """
    return html_utils.html_to_text(document_data)


def init_textract():
    """Make sure textract is loaded"""
    import textract as ex                # pylint: disable=import-outside-toplevel, import-error
    global textract
    textract = ex


def document_to_text(doc_filename):
    """Returns text version of document FILENAME of unspecified type"""
    text = ""
    try:
        init_textract()
        ## OLD: text = system.from_utf8(textract.process(doc_filename))
        text = textract.process(doc_filename)
        if isinstance(text, bytes):
            text = text.decode("UTF-8", errors="ignore")
    except:
        debug.trace_fmtd(3, "Warning: problem converting document file {f}: {e}",
                         f=doc_filename, e=sys.exc_info())
    debug.trace_fmt(5, "document_to_text({fn}) => {r}",
                    fn=doc_filename, r=gh.elide(text))
    return text


def extract_html_images(document_data=None, url=None, filename=None):
    """Returns list of all images in HTML DOC from URL (n.b., URL used to determine base URL)
    Warning: deprecated function; import from html_utils instead
    """
    return html_utils.extract_html_images(document_data, url, filename)


def version_to_number(version, max_padding=3):
    """Converts VERSION to number that can be used in comparisons
    Note: The Result will be of the form M.mmmrrrooo..., where M is the
    major number m is the minor, r is the revision and o is other.
    Each version component will be prepended with up MAX_PADDING [3] 0's
    Notes:
    - comparisons should be made against using functional constant, not number:
      EX: version_to_number("1.1.1") > version_to_number("1.1")
      EX: version_to_number("1.1.1") < 1.1
    - strings in the version are ignored
    - 0 is returned if version string is non-standard"""
    # EX: version_to_number("1.11.1") => 1.00110001
    # EX: version_to_number("1") => 1.0
    # EX: version_to_number("") => 0
    # TODO: support string (e.g., 1.11.2a).
    version_number = 0
    version_text = version
    new_version_text = ""
    max_component_length = (1 + max_padding)
    debug.trace_fmt(5, "version_to_number({v})", v=version)

    # HACK: Remove newlines
    # example: "3.7.6 (default, Jan  8 2020, 19:59:22)<NL>[GCC 7.3.0]"
    if my_re.search(r"^([^\n]+)\n", version_text):
        debug.trace_fmt(5, "Removing pesky newline: '{vt}' => '{nvt}'", vt=version_text, nvt=my_re.group(1))
        version_text = my_re.group(1)

    # Remove all non-numeric components, after initial number-like value
    # ex: "3.7.6 (default, Jan  8 2020, 19:59:22" => "3.7.6"
    ## DEBUG: debug.trace_fmt(6, "stripped sys.version: {v}", v=re.sub(r"[^0-9\.]", "", sys.version))
    ## BAD: version_text = re.sub(r"[^0-9\.]", "", version_text)
    version_text = re.sub(r"^([0-9\.]+)[^0-9\.].*", r"\1", version_text, flags=re.DOTALL)
    if (version_text != version):
        debug.trace_fmt(3, "Warning: stripped non-numeric suffix from version: '{v}' => '{nv}'", 
                        v=version, nv=version_text)
    debug.assertion(not re.search(r"[^0-9\.]", version_text))

    # Convert component numbers iteratively and zero-pad if necessary
    # NOTE: Components greater than max-padding + 1 treated as all 9's.
    debug.trace_fmt(4, "version_text: '{vt}'", vt=version_text)
    first = True
    num_components = 0
    ## BAD: regex = r"^(\d+)(\.((\d*).*))?$"
    regex = r"^(\d+)\.?((\d*).*)?$"
    while (my_re.search(regex, version_text)):
        component = my_re.group(1)
        # TODO: fix my_re.group to handle None as ""
        version_text = my_re.group(2) if my_re.group(2) else ""
        num_components += 1
        debug.trace_fmt(4, "remaining version_text: {vt}", vt=version_text)

        component = system.to_string(system.to_int(component))
        if first:
            new_version_text = component + "."
            ## OLD: regex = r"^(\d+)\.?((\d*).*)$"
            first = False
        else:
            # If component exceeds max length, replace with max value (e.g., "12345" => "9999")
            if (len(component) > max_component_length):
                old_component = component
                component = "9" * max_component_length
                debug.trace_fmt(2, "Warning: replaced overly long component #{n} '{oc}' with '{nc}'",
                                n=num_components, oc=old_component, nc=component)
            # Otherwise, zero-fill the component text (e.g., "12" => "0012")
            else:
                component = "0" * (max_component_length - len(component)) + component
            new_version_text += component
        debug.trace_fmt(4, "Component {n}: '{c}'", n=num_components, c=component)
        debug.trace_fmt(4, "new_version_text: '{nvt}'", nvt=new_version_text)
    debug.assertion(not re.search(r"^[0-9\.]", version_text))

    version_number = system.to_float(new_version_text, version_number)
    ## TODO:
    ## if (my_re.search(p"[a-z]", version_text, re.IGNORECASE)) {
    ##     version_text = my_re.... 
    ## }
    debug.trace_fmt(4, "version_to_number({v}) => {n}", v=version, n=version_number)
    return version_number


def extract_string_list(text):
    """Extract list of string values in TEXT string separated by spacing or a comma.
    Note: quotes can be used if there are embedded spaces"""
    ## OLD: Note: the string values currently cannot be quoted (i.e., no embedded spaces)."""
    # EX: extract_string_list("1, 2,3 4") => ['1', '2', '3', '4']
    # EX: extract_string_list(" ") => []
    # EX: extract_string_list(",,") => []
    # EX: extract_string_list("'my dog'  likes  'my  hot  dog'") => ['my dog', 'likes', 'my  hot  dog']

    # Convert commas and whitespace to be blanks
    # Also removes surrounding list brackets
    normalized_text = text.replace(",", " ").strip()
    normalized_text = my_re.sub(r"^\[(.*)\]$", r"\1", normalized_text)
    normalized_text = re.sub(r"\s", " ", normalized_text)
    debug.trace_fmtd(6, "normalized_text1: {nt}", nt=normalized_text)

    # Change spaces within embedded quotes to be new-page (i.e., ^L)
    max_tries = 1 + text.count(" ")
    num_tries = 0
    quoted_space_regex = r"(['\"])([^ '\"]+) ([^'\"]+)\1"
    while (num_tries < max_tries):
        num_tries += 1
        new_normalized_text = re.sub(quoted_space_regex, r"\1\2\f\3\1", normalized_text)
        if (new_normalized_text == normalized_text):
            break
        normalized_text = new_normalized_text
        debug.trace_fmtd(7, "sub{i}: {nt}", i=num_tries, nt=normalized_text)
    debug.assertion(num_tries <= max_tries)
    debug.trace_fmtd(6, "num_tries={nt}; max_tries={mt}", nt=num_tries, mt=max_tries)
    debug.trace_fmtd(6, "normalized_text2: {nt}", nt=normalized_text)

    # Change other spaces to be tab delimited then restore spaces.
    normalized_text = re.sub(" +", "\t", normalized_text)
    normalized_text = re.sub("\f", " ", normalized_text)
    debug.trace_fmtd(6, "normalized_text3: {nt}", nt=normalized_text)
    
    # Split based on tabs and remove outer quotes
    ## OLD: value_list = re.split(" +", normalized_text)
    value_list = re.split("\t", normalized_text)
    if (value_list == [""]):
        value_list = []
    for (i, value) in enumerate(value_list):
        if re.search("^" + quoted_space_regex + "$", value):
           value_list[i] = value[1:-1]
    debug.assertion("" not in value_list)
    debug.trace_fmtd(5, "extract_string_list({t}) => {vl}", t=text, vl=value_list)
    return value_list


def extract_int_list(text, default_value=0):
    """Extract list of integral values from comma and/or whitespace delimited TEXT using DEFAULT_VALUE for non-integers (even if floating point)"""
    return [to_int(v, default_value) for v in extract_string_list(text)]


def getenv_ints(var, default_values_spec):
    """Get integer list using values specified for environment VAR (or DEFAULT_VALUES_SPEC text)"""
    # EX: getenv_ints("DUMMY VARIABLE", str(list(range(5)))) => [0, 1, 2, 3, 4]
    return extract_int_list(system.getenv_text(var, default_values_spec))


def is_symbolic(token):
    """Indicates whether (string) token is symbolic (e.g., non-numeric like "ABC").
    Note: for convenience, tokens can be numeric types (e.g., 17), with False returned."""
    # EX: is_symbolic("PI") => True
    # EX: is_symbolic("3.14159") => False
    # EX: is_symbolic("123") => False
    # EX: is_symbolic(123) => False
    # EX: is_symbolic(0.1) => False
    # TODO: add support for complex numbers
    in_token = token
    result = True
    try:
        if isinstance(token, six.string_types):
            token = token.strip()
        # Note: if an exception is not triggered, the token is numeric
        if (float(token) or int(token)):
            debug.trace_fmt(6, "is_symbolic: '{t}' is numeric", t=token)
        result = False
    except (TypeError, ValueError):
        pass
    debug.trace_fmt(7, "is_symbolic({t}) => {r})", t=in_token, r=result)
    return result


def make_fixed_length(text, length=16):
    """Return TEXT with padding up to LENGTH chars; similar to str.ljust except for trucation"""
    # ex: make_fixed_length("fubar", 3) => "fub"
    # ex: make_fixed_length("fub", 5) => "fub  "
    result = text.ljust(length)[:length]
    debug.trace_fmt(7, "make_fixed_length({t}, [len={l}]) => {r}",
                    t=text, l=length, r=result)
    return result

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    system.print_stderr("Warning: not intended for command-line use")
    debug.assertion("html" not in html_to_text("<html><body></body></html>"))



#! /usr/bin/env python3
#
# Test(s) for ../check_time_tracking.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_check_time_tracking.py
#

"""Tests for check_time_tracking module"""

# Standard modules
## TODO: from collections import defaultdict

# Installed modules
import pytest

# Local modules
from mezcla.unittest_wrapper import TestWrapper
## TODO: from mezcla.unittest_wrapper import trap_exception
from mezcla import debug
## TODO: from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re
from mezcla import system

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                        global module object
#    TestIt.script_module:              path to file
try:
    import mezcla.adhoc.check_time_tracking as THE_MODULE
except:
    debug.assertion(False, "TODO1: FIXME")
    THE_MODULE = None
debug.trace_expr(5, THE_MODULE)
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

TEMPLATE = """
    # Time tracking for work Jan 2222
    # John J. Doe
    #
    # legend:
    # XYZ: XYZ, Inc.
    # JJD: personal stuff
    #
    
    --------------------------------------------------------------------------------
    
    Week of Sunday DD MMM YYYY
    
    Sun:            # dd mmm
    Hours:
    
    Mon:
    # <here>
    Hours:
    
    Tues:
    Hours:
    
    Wed:
    12-1: [JJD] break
    Hours:
    
    Thurs:
    Hours:
    
    Fri:
    Hours:
    
    Sat:
    Hours:
    
    # TODO
    # Misc: task1, task2, ..., task3
    # Hours:
    
    Weekly hours:
    
    --------------------------------------------------------------------------------
    
    Total hours:
"""

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    #
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_01_data_file(self):
        """Tests run_script w/ data file"""
        debug.trace(4, f"TestIt.test_01_data_file(); self={self}")
        system.write_file(self.temp_file, TEMPLATE)
        output = self.run_script(options="", data_file=self.temp_file)
        self.do_assert(my_re.search(r"Total hours: 0", output.strip()))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    ## DEBUG: @trap_exception            # TODO: remove when debugged
    def test_02_something_else(self):
        """Test for something_else: TODO..."""
        debug.trace(4, f"TestIt.test_02_something_else(); self={self}")
        self.do_assert(False)
        ## ex: self.do_assert(THE_MODULE.TODO_function() == TODO_value)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_03_tagged_hours(self):
        """Test printed report for tags (capsys-like)"""
        debug.trace(4, f"TestIt2.test_03_tagged_hours(); self={self}")
        tag_hash = {"JJD": 1.0}
        THE_MODULE.show_tagged_hours(tag_hash)
        captured = self.get_stdout()
        self.do_assert("Hours by tag.*JJD" in captured)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    ## TODO2: here and elsewhere: invoke_tests(__file__)
    pytest.main([__file__])



#! /usr/bin/env python3
#
# Test(s) for ../html_utils.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - TODO: If any of the setup/cleanup methods defined, make sure to invoke base
#   (see examples below for setUp and tearDown).
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_html_utils.py
#

"""Tests for html_utils module"""

# Standard packages
import re
import os

# Installed packages
import pytest
## OLD: import bs4

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
from mezcla import system
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
## TODO: template => new name
import mezcla.html_utils as THE_MODULE

# Environment options
TEST_SELENIUM = system.getenv_bool("TEST_SELENIUM", False,
                                   "Include tests requiring selenium")

class TestHtmlUtils(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    scrappycito_url = "http://www.scrappycito.com"
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_browser(self):
        """Ensure get_browser() works as expected"""
        debug.trace(4, "test_get_browser()")
        assert False, "TODO: code test"

    def test_get_url_parameter_value(self):
        """Ensure get_url_parameter_value works as expected"""
        debug.trace(4, "test_get_url_parameter_value()")
        save_user_parameters = THE_MODULE.user_parameters
        THE_MODULE.user_parameters = {}
        assert THE_MODULE.get_url_parameter_value("fubar", None) is None
        assert THE_MODULE.get_url_parameter_value("fubar", None, {"fubar": "fu"}) == "fu"
        THE_MODULE.user_parameters = {"fubar": "bar"}
        assert THE_MODULE.get_url_parameter_value("fubar", None) == "bar"
        assert THE_MODULE.get_url_parameter_value("fubar", None, {"fubar": "fu"}) == "fu"
        THE_MODULE.user_parameters = save_user_parameters
        return

    def test_get_inner_text(self):
        """Verify that JavaScript fills in window dimensions
        Note: requires selenium"""
        debug.trace(4, "test_get_inner_text()")
        if not TEST_SELENIUM:
            debug.trace(4, "Ignoring test_get_inner_text as selenium required")
            return
        html_filename = "simple-window-dimensions.html"
        html_path = gh.resolve_path(html_filename)
        url = ("file:" + system.absolute_path(html_path))
        # TODO: use direct API call to return unrendered text
        unrendered_text = gh.run(f"lynx -dump {url}")
        debug.trace_expr(5, unrendered_text)
        assert re.search(r"Browser dimensions: \?", unrendered_text)
        rendered_text = THE_MODULE.get_inner_text(url)
        debug.trace_expr(5, rendered_text)
        assert re.search(r"Browser dimensions: \d+x\d+", rendered_text)

    def test_get_inner_html(self):
        """Verify that JavaScript fills in window dimensions
        Note: requires selenium"""
        debug.trace(4, "test_get_inner_html()")
        if not TEST_SELENIUM:
            debug.trace(4, "Ignoring test_get_inner_html as selenium required")
            return
        html_filename = "simple-window-dimensions.html"
        html_path = gh.resolve_path(html_filename)
        url = ("file:" + system.absolute_path(html_path))
        # TODO: use direct API call to return unrendered text
        unrendered_html = gh.run(f"lynx -source {url}")
        debug.trace_expr(5, unrendered_html)
        assert re.search(
            r"<li>Browser dimensions:\s*<span.*>\?\?\?</span></li>",
            unrendered_html,
            )
        rendered_html = THE_MODULE.get_inner_html(url)
        debug.trace_expr(5, rendered_html)
        assert re.search(
            r"<li>Browser dimensions:\s*<span.*>\d+x\d+</span></li>",
            rendered_html,
            )

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_inner_html_alt(self):
        """Alternative test of get_inner_html"""
        debug.trace(4, f"TestIt.test_get_inner_html_alt(); self={self}")
        output = THE_MODULE.get_inner_html(self.scrappycito_url)
        self.do_assert(not my_re.search(r"sign.out", output.strip(), flags=my_re.IGNORECASE))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_run_script_for_inner_html(self):
        """Test of getting inner HTML via script invocation"""
        debug.trace(4, f"TestIt.test_run_script_for_inner_html(); self={self}")
        output = self.run_script(options=f"--inner --stdout {self.scrappycito_url}", data_file=self.temp_file)
        self.do_assert(not my_re.search(r"sign.out", output.strip(), flags=my_re.IGNORECASE))
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_document_ready(self):
        """Ensure document_ready() works as expected"""
        debug.trace(4, "test_document_ready()")
        assert False, "TODO: code test"

    def test_escape_html_value(self):
        """Ensure escape_html_value() works as expected"""
        debug.trace(4, "test_escape_html_value()")
        # note: this test is the same as system.test_escape_html_text
        assert THE_MODULE.escape_html_value("<2/") == "&lt;2/"
        assert THE_MODULE.escape_html_value("Joe's hat") == "Joe&#x27;s hat"

    def test_unescape_html_value(self):
        """Ensure unescape_html_value() works as expected"""
        debug.trace(4, "test_unescape_html_value()")
        # note: this test is the same as test_system.test_unescape_html_text
        assert THE_MODULE.unescape_html_value("&lt;2/") == "<2/"
        assert THE_MODULE.unescape_html_value("Joe&#x27;s hat") == "Joe's hat"

    def test_escape_hash_value(self):
        """Ensure escape_hash_value() works as expected"""
        debug.trace(4, "test_escape_hash_value()")
        hash_table = {
            'content-type': 'multipart/form-data;\n',
            'name': 'description'
        }
        assert THE_MODULE.escape_hash_value(hash_table, 'content-type') == 'multipart/form-data;<br>'

    def test_get_param_dict(self):
        """Ensure get_param_dict() works as expected"""
        debug.trace(4, "test_get_param_dict()")
        THE_MODULE.user_parameters = {
            'https-port': '443',
            'not-found-status': '404',
            'redirect-status': '302'
        }
        assert THE_MODULE.get_param_dict('not-found-status') == 'not-found-status'
        assert THE_MODULE.get_param_dict() == THE_MODULE.user_parameters

    def test_set_param_dict(self):
        """Ensure set_param_dict() works as expected"""
        debug.trace(4, "test_set_param_dict()")
        THE_MODULE.user_parameters = {
            'not-found-status': '404',
            'redirect-status': '302'
        }
        new_user_parameters = {'https-port': '443'}
        THE_MODULE.issued_param_dict_warning = False
        THE_MODULE.set_param_dict(new_user_parameters)
        assert THE_MODULE.user_parameters == new_user_parameters
        assert THE_MODULE.issued_param_dict_warning

    def test_get_url_param(self):
        """Ensure get_url_param() works as expected"""
        debug.trace(4, "test_get_url_param()")
        THE_MODULE.user_parameters = {
            'not-found-status': '404',
            'redirect-status': '302',
            'default-body': "Joe's hat"
        }
        assert THE_MODULE.get_url_param('redirect-status') == '302'
        assert THE_MODULE.get_url_param('bad-request-status', default_value='400') == '400'
        assert THE_MODULE.get_url_param('default-body', escaped=True) == 'Joe&#x27;s hat'

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_url_param_checkbox_spec(self):
        """Ensure get_url_param_checkbox_spec() works as expected"""
        debug.trace(4, "test_get_url_param_checkbox_spec()")
        param_dict = {"check_1": "on", "check_2": "off","check_3": "True",
                                    "check_4": False, "check_5": 1}

        # Test multiple positive cases
        assert THE_MODULE.get_url_param_checkbox_spec("check_1", param_dict=param_dict)
        assert THE_MODULE.get_url_param_checkbox_spec("check_3", param_dict=param_dict)
        assert THE_MODULE.get_url_param_checkbox_spec("check_5", param_dict=param_dict)

        # test non-checked and non-existent check cases
        assert not THE_MODULE.get_url_param_checkbox_spec("check_2", param_dict=param_dict)
        assert not THE_MODULE.get_url_param_checkbox_spec("check_4", param_dict=param_dict)
        assert not THE_MODULE.get_url_param_checkbox_spec("non_check", param_dict=param_dict)

    def test_get_url_parameter_bool(self):
        """Ensure get_url_parameter_bool() works as expected"""
        debug.trace(4, "test_get_url_parameter_bool()")
        assert THE_MODULE.get_url_parameter_bool("abc", False, { "abc": "on" })
        assert THE_MODULE.get_url_param_bool("abc", False, { "abc": "True" })

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_get_url_parameter_int(self):
        """Ensure get_url_parameter_int() works as expected"""
        debug.trace(4, "test_get_url_parameter_int()")
        assert THE_MODULE.get_url_parameter_int("abc", 0, { "abc": "123" }) == 123
        assert THE_MODULE.get_url_parameter_int("abc", 0, { "abc": "123.4" }) == 0
        assert THE_MODULE.get_url_parameter_int("abc", 0, { "abc": "not int" }) == 0

    def test_fix_url_parameters(self):
        """Ensure fix_url_parameters() works as expected"""
        debug.trace(4, "test_fix_url_parameters()")
        assert THE_MODULE.fix_url_parameters({'w_v':[7, 8], 'h_v':10}) == {'w-v':8, 'h-v':10}

    def test_expand_misc_param(self):
        """Ensure expand_misc_param() works as expected"""
        debug.trace(4, "test_expand_misc_param()")
        misc_dict = {
            'x': 1,
            'y': 2,
            'z': 'a=3, b=4',
        }
        expected = {
            'x': 1,
            'y': 2,
            'z': 'a=3, b=4',
            'a': '3',
            'b': '4',
        }
        assert THE_MODULE.expand_misc_param(misc_dict, 'z') == expected

    @pytest.mark.xfail                   # TODO: remove xfail
    def test__read_file(self):
        """Ensure _read_file() works as expected"""
        debug.trace(4, "test__read_file()")

        # test valid file
        temp_file = self.get_temp_file()
        system.write_file(temp_file, 'file\nwith\nmultiple\nlines\n')
        assert (
            THE_MODULE._read_file(filename=temp_file, as_binary=False) ==
            'file\nwith\nmultiple\nlines\n')

        # Test invalid file
        debug.set_level(3)
        THE_MODULE._read_file(filename='invalid_file', as_binary=False)
        captured = self.get_stderr()
        assert "Unable to read file" in captured

        # Test binary mode
        test_filename = self.create_temp_file("open binary")
        assert (
            THE_MODULE._read_file(filename=test_filename, as_binary=True) ==
            bytes("open binary"+ os.linesep , "UTF-8"))

    @pytest.mark.xfail                   # TODO: remove xfail
    def test__write_file(self):
        """Ensure _write_file() works as expected"""
        debug.trace(4, "test__write_file()")
        # Test normal usage
        filename = self.get_temp_file()
        THE_MODULE._write_file(filename, "it", as_binary=False)
        assert THE_MODULE._read_file(filename=filename, as_binary=False) == "it\n"

        # Test binary mode
        filename = self.get_temp_file()
        THE_MODULE._write_file(filename, data=bytes("it", encoding="UTF-8"), as_binary=True)
        assert THE_MODULE._read_file(filename=filename, as_binary=True) == b"it"

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_old_download_web_document(self):
        """Ensure old_download_web_document() works as expected"""
        ## NOTE: Currently fails if UTF-8 error occurs
        debug.trace(4, "test_old_download_web_document()")
        assert "<!doctype html>" in THE_MODULE.old_download_web_document("https://www.google.com")

    def test_download_web_document(self):
        """Ensure download_web_document() works as expected"""
        debug.trace(4, "test_download_web_document()")
        assert "currency" in THE_MODULE.download_web_document("https://simple.wikipedia.org/wiki/Dollar")
        assert THE_MODULE.download_web_document("www. bogus. url.html") is None

    def test_test_download_html_document(self):
        """Ensure test_download_html_document() works as expected"""
        debug.trace(4, "test_test_download_html_document()")
        assert "Google" in THE_MODULE.test_download_html_document("www.google.com")
        assert "Tomás" not in THE_MODULE.test_download_html_document("http://www.tomasohara.trade", encoding="big5")

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_download_html_document(self):
        """Ensure download_html_document() works as expected"""
        debug.trace(4, "test_download_html_document()")

        # Set tmp_dir and filename for testing
        tmp_dir = system.getenv("TMP")
        filename = "test_download_file"

        # Assert file is downloaded and created in tmp_dir
        THE_MODULE.download_html_document("http://www.tomasohara.trade", download_dir=tmp_dir, filename=filename)
        assert filename in system.read_directory(tmp_dir)

        # Assert exception report is printed when not Ignore
        try :
            _ = THE_MODULE.download_html_document("", ignore=False)
        except Exception as _:
            pass
        err = self.get_stderr()
        assert "Error during retrieve_web_document" in err

        # Assert exception report is not printed when Ignore
        self.clear_stderr()
        try :
            _ = THE_MODULE.download_html_document("", ignore=True)
        except Exception as _:
            pass
        err = self.get_stderr()
        assert "Error during retrieve_web_document" not in err




    @pytest.mark.xfail                   # TODO: remove xfail
    def test_download_binary_file(self):
        """Ensure download_binary_file() works as expected"""
        debug.trace(4, "test_download_binary_file()")
        binary_doc = THE_MODULE.download_binary_file(url="www.tomasohara.trade")
        non_binary_doc = THE_MODULE.download_web_document(url="www.tomasohara.trade")
        assert re.search(b"Scrappy.*Cito", binary_doc)
        assert bytes(non_binary_doc, encoding="UTF-8") == binary_doc


    def test_retrieve_web_document(self):
        """Ensure retrieve_web_document() works as expected"""
        debug.trace(4, "test_retrieve_web_document()")
        assert re.search("Scrappy.*Cito", THE_MODULE.retrieve_web_document("www.tomasohara.trade"))

    def test_init_BeautifulSoup(self):
        """Ensure init_BeautifulSoup() works as expected"""
        debug.trace(4, "test_init_BeautifulSoup()")
        THE_MODULE.BeautifulSoup = None
        THE_MODULE.init_BeautifulSoup()
        assert THE_MODULE.BeautifulSoup

    def test_extract_html_link(self):
        """Ensure extract_html_link() works as expected"""
        debug.trace(4, "test_extract_html_link()")

        html = (
            '<!DOCTYPE html>\n'
            '<html>\n'
            '<body>\n'
            '<h2>The target Attribute</h2>\n'
            '<div class="some-class">this is a div</div>\n'
            '<div class="some-class another-class">'
            '<a href="https://www.anothersite.io/" target="_blank">another site</a>\n'
            '<a href="https://www.example.com/" target="_blank">Visit Example!</a>\n'
            '<a href="https://www.example.com/sopport" target="_blank">example sopport</a>\n'
            '</div>'
            '<a href="https://www.subdomain.example.com/" target="_blank">Visit subdomain of Example!</a>\n'
            '<a href="https://www.example.com.br/" target="_blank">visit Example Brazil!</a>\n'
            '<p>If target="_blank", this is a link.</p>\n'
            '<a href="www.subdomain.example.com/sitemap.xml" target="_blank">see the sitemap</a>\n'
            '<a href="/home.html" target="_blank">home page</a>\n'
            '</body>\n'
            '</html>\n'
        )

        # NOTE that the last two urls has a extra '/'.
        ## TODO: check if the extra '/' in the last two urls are correct.
        all_urls = [
            'https://www.anothersite.io/',
            'https://www.example.com/',
            'https://www.example.com/sopport',
            'https://www.subdomain.example.com/',
            'https://www.example.com.br/',
            'http:///www.subdomain.example.com/sitemap.xml',
            'https://www.example.com//home.html',
        ]

        # Test extract all urls from html
        assert THE_MODULE.extract_html_link(html, url='https://www.example.com', base_url='http://') == all_urls

        # Test base_url none
        ## TODO: this assertion is returning, need to be solved:
        ##      https://www.example.com//www.subdomain.example.com/sitemap.xml
        ## assert THE_MODULE.extract_html_link(html, url='https://www.example.com') == all_urls

    @pytest.mark.xfail
    def test_get_inner_html_type_hints(self):
        URL_VALID = "https://duckduckgo.com"
        URL_INVALID = "duckduckgo.com"
        assert THE_MODULE.get_inner_html(URL_VALID)
        assert not THE_MODULE.get_inner_html(URL_INVALID)

    @pytest.mark.xfail
    def test_get_inner_text_alt(self):
        URL_VALID = "https://duckduckgo.com"
        URL_INVALID = "duckduckgo.com"
        inner_text_valid = THE_MODULE.get_inner_text(URL_VALID)
        inner_text_invalid = THE_MODULE.get_inner_text(URL_INVALID)
        keywords = ["DuckDuckGo", "private", "free", "browser", "search", "ads"]
        assert all(word in inner_text_valid for word in keywords)
        assert not inner_text_invalid

    @pytest.mark.xfail
    def test_document_ready_alt(self):
        URL_VALID = "https://duckduckgo.com"
        assert (THE_MODULE.document_ready(URL_VALID))
        assert not THE_MODULE.document_ready(self.scrappycito_url)
        ## Exception raised:
        # E       selenium.common.exceptions.JavascriptException: Message: TypeError: document.body is null
        # E       Stacktrace:
        # E       @http://www.scrappycito.com:9330/:2:7
        # E       @http://www.scrappycito.com:9330/:3:8
        # /home/zavee/.local/lib/python3.10/site-packages/selenium/webdriver/remote/errorhandler.py:229: JavascriptException

    @pytest.mark.xfail
    def test_set_param_dict_alt(self):
        param_dict = {
            1: "a+b+c",
            2: "a%2b%2c%2"
        }
        THE_MODULE.set_param_dict(param_dict)
        assert THE_MODULE.user_parameters == param_dict

    @pytest.mark.xfail
    def test_set_param_dict_type_hints(self):
        param_dict = str({1: "a+b+c", 2: "a%2b%2c%2"})
        THE_MODULE.set_param_dict(param_dict)
        assert not isinstance(THE_MODULE.user_parameters, dict)
        assert len(THE_MODULE.user_parameters) > 2

    @pytest.mark.xfail    
    def test_get_param_dict_type_hints(self):
        param_dict = str({
            "p1": "a+b+c",
            "p2": "a%2b%2c%2"
        })
        result = THE_MODULE.get_param_dict(param_dict=param_dict)
        assert not isinstance(result, dict)

    @pytest.mark.xfail 
    def test_get_url_param_type_hints(self):
        param_dict = {
            "name": "Terry",
            "age": "30"
        }
        name = "name"
        default_value = "Default"
        escaped = False
        result = THE_MODULE.get_url_param(name, default_value, param_dict, escaped)

        assert isinstance(name, str)
        assert isinstance(default_value, str) or default_value is None
        assert isinstance(param_dict, dict) or param_dict is None
        assert isinstance(escaped, bool)
        assert isinstance(result, str)
    
    @pytest.mark.xfail 
    def test_get_url_text_type_hints(self):
        param_dict = {
            "name": "Alice",
            "age": 28,
            "city": "New York",
            "is_student": False,
            "grades": [85, 90, 78, 92, 88]
        }

        name = "name"
        default_value = "default"
        
        result = THE_MODULE.get_url_text(name, default_value, param_dict)
        assert isinstance(result, str)
        assert isinstance(name, str)
        assert isinstance(default_value, str) or default_value is None
        assert isinstance(param_dict, dict) or param_dict is None

        result = THE_MODULE.get_url_text(None)
        assert isinstance(result, str)

    @pytest.mark.xfail 
    def test_get_url_param_checkbox_spec_type_hints(self):
        name = "param"
        default_value = ""
        param_dict = {name: "on"}

        result = THE_MODULE.get_url_param_checkbox_spec(name, default_value, param_dict)

        assert isinstance(name, str)
        assert isinstance(default_value, (bool, str)) or default_value is None
        assert isinstance(param_dict, dict) or param_dict is None
        assert isinstance(result, str)  # Result should always be a string

        result = THE_MODULE.get_url_param_checkbox_spec(None)
        assert isinstance(result, str)

    @pytest.mark.xfail 
    def test_get_url_parameter_value_type_hints(self):
        param = "param"
        default_value = ""
        param_dict = {param: ["on", "off", "neutral"]}
        
        result = THE_MODULE.get_url_parameter_value(param, default_value, param_dict)

        assert isinstance(param, str)
        assert isinstance(default_value, (str, type(None)))  # Union of Any and None
        assert isinstance(param_dict, dict) or param_dict is None
        assert isinstance(result, (str, str))  # Union of string and Any type

        result = THE_MODULE.get_url_parameter_value(None)
        assert result is None

    @pytest.mark.xfail 
    def test_get_url_parameter_bool_type_hints(self):
        param = "param"
        default_value = False
        param_dict = {param: "on"}
        
        result = THE_MODULE.get_url_parameter_bool(param, default_value, param_dict)

        assert isinstance(param, str)
        assert isinstance(default_value, bool)
        assert isinstance(param_dict, dict) or param_dict is None
        assert not isinstance(result, str)
        assert isinstance(result, bool)

        result = THE_MODULE.get_url_parameter_bool(None)
        assert isinstance(result, bool)

    @pytest.mark.xfail 
    def test_get_url_parameter_int_type_hints(self):
        param = "param"
        default_value = 0
        param_dict = {param: 9}

        result = THE_MODULE.get_url_parameter_int(param, default_value, param_dict)

        assert isinstance(param, str)
        assert isinstance(default_value, int)
        assert isinstance(param_dict, dict) or param_dict is None
        assert isinstance(result, int)

        result = THE_MODULE.get_url_parameter_int(None)
        assert result is not None and isinstance(result, int)

    @pytest.mark.xfail 
    def test_get_url_parameter_float_type_hints(self):
        param = "param"
        default_value = 0.0
        param_dict = {param: 17.38}

        result = THE_MODULE.get_url_parameter_float(param, default_value, param_dict)

        assert isinstance(param, str)
        assert isinstance(default_value, float)
        assert isinstance(param_dict, dict) or param_dict is None
        assert isinstance(result, float)

        result = THE_MODULE.get_url_parameter_float(None)
        assert result is not None and isinstance(result, float)

    @pytest.mark.xfail 
    def test_fix_url_parameters_type_hints(self):
        param = "param"
        param_dict = {param: [17.38, 19.45, 88.88, 16.09]}

        result = THE_MODULE.fix_url_parameters(param_dict)

        assert isinstance(param_dict, dict)
        assert isinstance(result, dict)
        assert not isinstance(result[param], list)  

        result = THE_MODULE.fix_url_parameters(None)
        assert result is not None and isinstance(result, float)

    @pytest.mark.xfail 
    def test_expand_misc_param_type_hints(self):
        misc_dict = {'x': 1, 'y': 2, 'z': 'a=3, b=4'}
        param_name = "z"
        param_dict = {"a":"3", "b":"4"}

        result = THE_MODULE.expand_misc_param(misc_dict, param_name, param_dict)

        assert isinstance(misc_dict, dict)
        assert isinstance(param_dict, dict) or param_dict is None
        assert isinstance(param_name, str)
        assert isinstance(result, dict)

        result = THE_MODULE.expand_misc_param(None, None)
        assert isinstance(result, dict) or result is None
        
    @pytest.mark.xfail
    def test__read_file_type_hints(self):
        contents = "Hello World"
        filename = self.create_temp_file(contents)
        as_binary = False
        result = THE_MODULE._read_file(filename, as_binary)

        assert isinstance(filename, str)
        assert isinstance(as_binary, bool)
        assert isinstance(result, str)        

    @pytest.mark.xfail
    def test__write_file_type_hints(self):
        data = "Hello World"
        filename = self.create_temp_file(contents="")
        as_binary = False
        result = THE_MODULE._write_file(filename, data, as_binary)

        assert isinstance(data, (str, bytes))
        assert isinstance(filename, str)
        assert isinstance(as_binary, bool)
        assert result is None  
        
    # @pytest.mark.xfail
    # def test_download_web_document_type_hints(self):

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Functions for operating system related access, such as running command or
# getting environment values.
#
# TODO:
# - Move functions from system.py here.
#

"""OS utilities: mainly wrappers around os package"""

# Standard modules
import os

# Local modules
from mezcla import debug
from mezcla import system

# Constants
TL = debug.TL


def split_extension(path):
    """Returns basename and extension for PATH"""
    result = os.path.splitext(path)
    try:
        filename_proper, ext = result
        debug.assertion(system.remove_extension(path) == filename_proper)
        debug.assertion(system.get_extension(path) == ext[1:])
    except:
        pass
    debug.trace(5, f"split_extension({path}) => {result}")
    return result

## TEMP: define dummy function for tests/test_os_utils.py
##
## def some_other_function():
##     """Used for testing test_02_no_other_functions"""
##     return

def main(*args, **kwargs):
    """Supporting code for command-line processing"""
    debug.trace_fmtd(6, "main{a}; kw=", a=args, kw=kwargs)
    system.print_stderr("Warning: {f} not intended for direct invocation!",
                        f=system.filename_proper(__file__))
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    debug.trace(5, f"module __doc__: {__doc__}")
    debug.assertion("TODO:" not in __doc__)
    main()



#! /usr/bin/env python3
#
# Simple tensorflow matrix multiplication based on following:
#    https://www.tensorflow.org/guide/gpu
#
# Sample usages:
# - Using defaults
#   NUM_ROWS=5000 NUM_COLS=5000 python tensorflow_matrix_multiply.py
#
# - Alternative (using environment variables for GPU usage):
#   export TF_FORCE_GPU_ALLOW_GROWTH=true
#   export NVIDIA_VISIBLE_DEVICES="0"
#   export CUDA_VISIBLE_DEVICES="0"
#   python tensorflow_matrix_multiply.py
#

"""Tensorflow matrix multiplication (for GPU usage diagnosis)"""

# Standard packages
import os
import random

# Installed packages
import tensorflow as tf

# Local pakcages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system
from mezcla import tpo_common as tpo

# Environment options
NUM_ROWS = system.getenv_int("NUM_ROWS", 1000)
NUM_COLS = system.getenv_int("NUM_COLS", 1000)
DEVICE_ID = system.getenv_value("DEVICE_ID", None)
ALLOW_GROWTH = system.getenv_bool("ALLOW_GROWTH", True)

def random_vector(size):
    """Returns random list of SIZE floats"""
    return [random.random() for _i in range(size)]

def random_row(size): 
    """Returns row (list) of SIZE floats"""
    return random_vector(size)

def random_matrix(num_rows, num_cols):
    """Returns matrix of size NUM_ROWS x NUM_COLS filled with random float in [0, 1)"""
    return [random_row(num_cols) for _r in range(num_rows)]

def tensor_multiply(a, b):
    """Compute matrix multiplication via tensorflow"""
    c = tf.matmul(a, b)
    debug.trace(6, f"A: {a}\nB: {b}\nC: {c}")
    #
    session = tf.compat.v1.Session()
    result = session.run(c)
    debug.trace(6, f"tensor_multiply() => {result}")
    #
    return result


def main():
    """Multiply two matrices on the GPU"""

    # Show simple usage if --help given
    dummy_app = Main(description=__doc__, skip_input=False, manual_input=False)
    
    # Show final NVidia GPU usage
    if dummy_app.verbose:
        print("initial GPU usage:\n", gh.run("nvidia-smi"))

    # note: need to use version-specific stuff for basic functions
    # pylint: disable=no-member
    if (tf.__version__[0] >= '2'):
        tf.compat.v1.disable_eager_execution()
        tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.DEBUG)
    else:
        tf.logging.set_verbosity(tf.logging.DEBUG)
    tf.debugging.set_log_device_placement(True)
    device_id = DEVICE_ID
    debug.assertion((device_id is None) or tpo.is_numeric(device_id))
    debug.trace(4, f"using device {device_id}")
    if ALLOW_GROWTH:
        system.setenv("TF_FORCE_GPU_ALLOW_GROWTH", "true")
    if device_id:
        system.setenv("NVIDIA_VISIBLE_DEVICES", device_id)
    debug.trace(5, f"environment: {{\n{os.environ}\n}}")
    a = tf.constant(random_matrix(NUM_ROWS, NUM_COLS))
    b = tf.constant(random_matrix(NUM_ROWS, NUM_COLS))
    result = tensor_multiply(a, b)
    print(f"A x B: \n{result}")

    # Show final NVidia GPU usage
    if dummy_app.verbose:
        print("final GPU usage:\n", gh.run("nvidia-smi"))

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python3
#
# Miscellaneous XML utility functions
#
#--------------------------------------------------------------------------------
# Notes:
# - The tail text is not intuitive, including what would seem to belong to the
#   parent node. Via https://docs.python.org/3/library/xml.etree.elementtree.html:
# 
#   <a>                                     a.text = a.tail = None
#     <b>1                  b.text
#       <c>2                c.text
#         <d/>
#         3                 d.tail
#       </c>
#     </b>
#     4                     b.tail
#   </a>
# 
# 
# text: text between element’s start tag and its first child or end tag
# tail: text between the element’s end tag and the next tag
# 

"""XML utility functions"""

# Standard packages
import sys
import xml.etree.ElementTree as ET

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system


def parse_xml(xml_text):
    """Parse XML_text, returning root node"""
    root = None
    try:
        tree = ET.ElementTree(ET.fromstring(xml_text))
        root = tree.getroot()
        debug.trace_object(5, root, "xml root")
    except:
        system.print_exception_info("parse_xml")
    debug.trace(4, f"parse_xml({xml_text}) => {root}")
    return root


def etree_to_dict(node):
    """Convert XML parse at NODE to dict"""
    debug.trace(5, f"etree_to_dict({node})")
    debug.assertion(node is not None)
    d = {node.tag : list(map(etree_to_dict, list(node)))}
    d.update(('@' + k, v) for k, v in node.attrib.items())
    d['text'] = node.text
    d['tail'] = node.tail
    d['inner'] = "".join(node.itertext())
    return d


def get_xml_text(node, depth=0):
    """Get all text for XML node at ROOT"""
    # TODO: get_xml_text => get_xml_tree_text???
    indent = "    " * depth if __debug__ else ""
    debug.trace(5, f"{indent}get_xml_text({node}, {depth})")
    debug.assertion(node is not None)
    debug.trace(4, f'[l] {node.tag + ": "}', no_eol=True)
    text = ("\t" * depth) + node.tag + ": "
    if node.text:
        debug.trace(4, f"[n] {node.text}", no_eol=True)
        text += node.text
    for child in node:
        text += "\n" + get_xml_text(child, 1 + depth)
    if node.tail:
        debug.trace(4, f"[t] {node.tail}", no_eol=True)
        text += "\n" + ("\t" * depth) + node.tail
    debug.trace(6, f"{indent}get_xml_text() => {text}")
    return text

#...............................................................................

def main(args):
    """Entry point for script"""
    xml_text = "<a><b>1<c>2<d/>3</c></b>4</a>"
    if (len(args) > 1):
        if (args[1] == "--help"):
            usage = (
                "\n"
                "usage: {prog} [--help] [xml-file|-]\n"
                "\n"
                "example: echo '<xml>How now<italics>brown cow</italics>?</xml>' | {prog} -\n").format(prog=args[0])
            system.exit(usage)
        elif (args[1] == "-"):
            xml_text = sys.stdin.read()
        else:
            xml_text = system.read_file(args[1])

    # Show input
    print("Input:")
    print(gh.indent(xml_text))
    print()

    # Show text as parse tree
    parsed_xml_text = get_xml_text(parse_xml(xml_text))
    debug.trace(4, f"\nlen={len(parsed_xml_text)}")
    print("XML parse tree")
    print(gh.indent_lines(parsed_xml_text))
    return

if __name__ == '__main__':
    ## OLD:
    ## system.print_error("Warning: Not intended for direct invocation.")
    ## system.print_error("A simple test follows.")
    main(sys.argv)
